@@ -44,8 +44,11 @@ to manage child processes.
 from __future__ import absolute_import
 
 import os
-import sys
-from gevent.hub import get_hub, reinit, PY3
+
+from gevent.hub import _get_hub_noargs as get_hub
+from gevent.hub import reinit
+from gevent._config import config
+from gevent._util import copy_globals
 import errno
 
 EAGAIN = getattr(errno, 'EAGAIN', 11)
@@ -70,54 +73,72 @@ if fcntl:
     __extensions__ += ['make_nonblocking', 'nb_read', 'nb_write']
 
     def make_nonblocking(fd):
-        """Put the file descriptor *fd* into non-blocking mode if possible.
+        """Put the file descriptor *fd* into non-blocking mode if
+        possible.
 
-        :return: A boolean value that evaluates to True if successful."""
+        :return: A boolean value that evaluates to True if successful.
+        """
         flags = fcntl.fcntl(fd, fcntl.F_GETFL, 0)
         if not bool(flags & os.O_NONBLOCK):
             fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
             return True
 
     def nb_read(fd, n):
-        """Read up to `n` bytes from file descriptor `fd`. Return a string
-        containing the bytes read. If end-of-file is reached, an empty string
-        is returned.
+        """
+        Read up to *n* bytes from file descriptor *fd*. Return a
+        byte string containing the bytes read, which may be shorter than
+        *n*. If end-of-file is reached, an empty string is returned.
 
         The descriptor must be in non-blocking mode.
         """
-        hub, event = None, None
-        while True:
-            try:
-                return _read(fd, n)
-            except OSError as e:
-                if e.errno not in ignored_errors:
-                    raise
-                if not PY3:
-                    sys.exc_clear()
-            if hub is None:
-                hub = get_hub()
-                event = hub.loop.io(fd, 1)
-            hub.wait(event)
+        hub = None
+        event = None
+        try:
+            while 1:
+                try:
+                    result = _read(fd, n)
+                    return result
+                except OSError as e:
+                    if e.errno not in ignored_errors:
+                        raise
+                if hub is None:
+                    hub = get_hub()
+                    event = hub.loop.io(fd, 1)
+                hub.wait(event)
+        finally:
+            if event is not None:
+                event.close()
+                event = None
+                hub = None
+
 
     def nb_write(fd, buf):
-        """Write bytes from buffer `buf` to file descriptor `fd`. Return the
-        number of bytes written.
+        """
+        Write some number of bytes from buffer *buf* to file
+        descriptor *fd*. Return the number of bytes written, which may
+        be less than the length of *buf*.
 
         The file descriptor must be in non-blocking mode.
         """
-        hub, event = None, None
-        while True:
-            try:
-                return _write(fd, buf)
-            except OSError as e:
-                if e.errno not in ignored_errors:
-                    raise
-                if not PY3:
-                    sys.exc_clear()
-            if hub is None:
-                hub = get_hub()
-                event = hub.loop.io(fd, 2)
-            hub.wait(event)
+        hub = None
+        event = None
+        try:
+            while 1:
+                try:
+                    result = _write(fd, buf)
+                    return result
+                except OSError as e:
+                    if e.errno not in ignored_errors:
+                        raise
+                if hub is None:
+                    hub = get_hub()
+                    event = hub.loop.io(fd, 2)
+                hub.wait(event)
+        finally:
+            if event is not None:
+                event.close()
+                event = None
+                hub = None
 
 
 def tp_read(fd, n):
@@ -140,6 +161,8 @@ def tp_write(fd, buf):
 
 
 if hasattr(os, 'fork'):
+    # pylint:disable=function-redefined,redefined-outer-name
+
     _raw_fork = os.fork
 
     def fork_gevent():
@@ -220,13 +243,16 @@ if hasattr(os, 'fork'):
             # XXX: Could handle tracing here by not stopping
             # until the pid is terminated
             watcher.stop()
-            _watched_children[watcher.pid] = (watcher.pid, watcher.rstatus, time.time())
-            if callback:
-                callback(watcher)
-            # dispatch an "event"; used by gevent.signal.signal
-            _on_child_hook()
-            # now is as good a time as any to reap children
-            _reap_children()
+            try:
+                _watched_children[watcher.pid] = (watcher.pid, watcher.rstatus, time.time())
+                if callback:
+                    callback(watcher)
+                # dispatch an "event"; used by gevent.signal.signal
+                _on_child_hook()
+                # now is as good a time as any to reap children
+                _reap_children()
+            finally:
+                watcher.close()
 
         def _reap_children(timeout=60):
             # Remove all the dead children that haven't been waited on
@@ -246,9 +272,11 @@ if hasattr(os, 'fork'):
             # just not waited on yet.
             now = time.time()
             oldest_allowed = now - timeout
-            dead = [pid for pid, val
-                    in _watched_children.items()
-                    if isinstance(val, tuple) and val[2] < oldest_allowed]
+            dead = [
+                pid for pid, val
+                in _watched_children.items()
+                if isinstance(val, tuple) and val[2] < oldest_allowed
+            ]
             for pid in dead:
                 del _watched_children[pid]
 
@@ -256,34 +284,73 @@ if hasattr(os, 'fork'):
             """
             Wait for a child process to finish.
 
-            If the child process was spawned using :func:`fork_and_watch`, then this
-            function behaves cooperatively. If not, it *may* have race conditions; see
+            If the child process was spawned using
+            :func:`fork_and_watch`, then this function behaves
+            cooperatively. If not, it *may* have race conditions; see
             :func:`fork_gevent` for more information.
 
-            The arguments are as for the underlying :func:`os.waitpid`. Some combinations
-            of *options* may not be supported (as of 1.1 that includes WUNTRACED).
+            The arguments are as for the underlying
+            :func:`os.waitpid`. Some combinations of *options* may not
+            be supported cooperatively (as of 1.1 that includes
+            WUNTRACED). Using a *pid* of 0 to request waiting on only processes
+            from the current process group is not cooperative. A *pid* of -1
+            to wait for any child is non-blocking, but may or may not
+            require a trip around the event loop, depending on whether any children
+            have already terminated but not been waited on.
 
             Availability: POSIX.
 
             .. versionadded:: 1.1b1
+            .. versionchanged:: 1.2a1
+               More cases are handled in a cooperative manner.
             """
+            # pylint: disable=too-many-return-statements
             # XXX Does not handle tracing children
+
+            # So long as libev's loop doesn't run, it's OK to add
+            # child watchers. The SIGCHLD handler only feeds events
+            # for the next iteration of the loop to handle. (And the
+            # signal handler itself is only called from the next loop
+            # iteration.)
+
             if pid <= 0:
                 # magic functions for multiple children.
                 if pid == -1:
-                    # Any child. If we have one that we're watching and that finished,
-                    # we need to use that one. Otherwise, let the OS take care of it.
+                    # Any child. If we have one that we're watching
+                    # and that finished, we will use that one,
+                    # preferring the oldest. Otherwise, let the OS
+                    # take care of it.
+                    finished_at = None
                     for k, v in _watched_children.items():
-                        if isinstance(v, tuple):
+                        if (
+                                isinstance(v, tuple)
+                                and (finished_at is None or v[2] < finished_at)
+                        ):
                             pid = k
-                            break
+                            finished_at = v[2]
+
                 if pid <= 0:
-                    # If we didn't find anything, go to the OS. Otherwise,
-                    # handle waiting
+                    # We didn't have one that was ready. If there are
+                    # no funky options set, and the pid was -1
+                    # (meaning any process, not 0, which means process
+                    # group--- libev doesn't know about process
+                    # groups) then we can use a child watcher of pid 0; otherwise,
+                    # pass through to the OS.
+                    if pid == -1 and options == 0:
+                        hub = get_hub()
+                        with hub.loop.child(0, False) as watcher:
+                            hub.wait(watcher)
+                            return watcher.rpid, watcher.rstatus
+                    # There were funky options/pid, so we must go to the OS.
                     return _waitpid(pid, options)
 
             if pid in _watched_children:
                 # yes, we're watching it
+
+                # Note that the remainder of this code must be careful to NOT
+                # yield to the event loop except at well known times, or
+                # we have a race condition between the _on_child callback and the
+                # code here that could lead to a process to hang.
                 if options & _WNOHANG or isinstance(_watched_children[pid], tuple):
                     # We're either asked not to block, or it already finished, in which
                     # case blocking doesn't matter
@@ -295,20 +362,42 @@ if hasattr(os, 'fork'):
                         return result[:2]
                     # it's not finished
                     return (0, 0)
-                else:
-                    # we should block. Let the underlying OS call block; it should
-                    # eventually die with OSError, depending on signal delivery
-                    try:
-                        return _waitpid(pid, options)
-                    except OSError:
-                        if pid in _watched_children and isinstance(_watched_children, tuple):
-                            result = _watched_children[pid]
-                            del _watched_children[pid]
-                            return result[:2]
-                        raise
+
+                # Ok, we need to "block". Do so via a watcher so that we're
+                # cooperative. We know it's our child, etc, so this should work.
+                watcher = _watched_children[pid]
+                # We can't start a watcher that's already started,
+                # so we can't reuse the existing watcher. Notice that the
+                # old watcher must not have fired already, or during this time, but
+                # only after we successfully `start()` the watcher. So this must
+                # not yield to the event loop.
+                with watcher.loop.child(pid, False) as new_watcher:
+                    get_hub().wait(new_watcher)
+                # Ok, so now the new watcher is done. That means
+                # the old watcher's callback (_on_child) should
+                # have fired, potentially taking this child out of
+                # _watched_children (but that could depend on how
+                # many callbacks there were to run, so use the
+                # watcher object directly; libev sets all the
+                # watchers at the same time).
+                return watcher.rpid, watcher.rstatus
+
+            # we're not watching it and it may not even  be our child,
+            # so we must go to the OS to be sure to get the right semantics (exception)
+            # XXX
+            # libuv has a race condition because the signal
+            # handler is a Python function, so the InterruptedError
+            # is raised before the signal handler runs and calls the
+            # child watcher
             # we're not watching it
             return _waitpid(pid, options)
 
+        def _watch_child(pid, callback=None, loop=None, ref=False):
+            loop = loop or get_hub().loop
+            watcher = loop.child(pid, ref=ref)
+            _watched_children[pid] = watcher
+            watcher.start(_on_child, watcher, callback)
+
         def fork_and_watch(callback=None, loop=None, ref=False, fork=fork_gevent):
             """
             Fork a child process and start a child watcher for it in the parent process.
@@ -337,10 +426,7 @@ if hasattr(os, 'fork'):
             pid = fork()
             if pid:
                 # parent
-                loop = loop or get_hub().loop
-                watcher = loop.child(pid, ref=ref)
-                _watched_children[pid] = watcher
-                watcher.start(_on_child, watcher, callback)
+                _watch_child(pid, callback, loop, ref)
             return pid
 
         __extensions__.append('fork_and_watch')
@@ -367,7 +453,7 @@ if hasattr(os, 'fork'):
             __extensions__.append('forkpty_and_watch')
 
         # Watch children by default
-        if not os.getenv('GEVENT_NOWAITPID'):
+        if not config.disable_watch_children:
             # Broken out into separate functions instead of simple name aliases
             # for documentation purposes.
             def fork(*args, **kwargs):
@@ -398,6 +484,23 @@ if hasattr(os, 'fork'):
                     # take any args to match fork_and_watch
                     return forkpty_and_watch(*args, **kwargs)
             __implements__.append("waitpid")
+
+            if hasattr(os, 'posix_spawn'):
+                _raw_posix_spawn = os.posix_spawn
+                _raw_posix_spawnp = os.posix_spawnp
+
+                def posix_spawn(*args, **kwargs):
+                    pid = _raw_posix_spawn(*args, **kwargs)
+                    _watch_child(pid)
+                    return pid
+
+                def posix_spawnp(*args, **kwargs):
+                    pid = _raw_posix_spawnp(*args, **kwargs)
+                    _watch_child(pid)
+                    return pid
+
+                __implements__.append("posix_spawn")
+                __implements__.append("posix_spawnp")
         else:
             def fork():
                 """
@@ -428,4 +531,8 @@ else:
     __implements__.remove('fork')
 
 
-__all__ = __implements__ + __extensions__
+__imports__ = copy_globals(os, globals(),
+                           names_to_ignore=__implements__ + __extensions__,
+                           dunder_names_to_keep=())
+
+__all__ = list(set(__implements__ + __extensions__))
